/*
 *   display.c
 *
 *   This file is part of Emu28
 *
 *   Copyright (C) 2002 Christoph Gieelink
 *
 */
#include "pch.h"
#include "resource.h"
#include "Emu28.h"
#include "ops.h"
#include "io.h"
#include "kml.h"

#define DISPLAY_FREQ	16					// display update 1/frequency (1/64) in ms

#define B 0x00000000						// black
#define W 0x00FFFFFF						// white
#define I 0xFFFFFFFF						// ignore

#define LCD_WIDTH	137
#define LCD_HEIGHT	32

#define LCD_ROW		(36*4)					// max. pixel per line

UINT    nBackgroundX = 0;
UINT    nBackgroundY = 0;
UINT    nBackgroundW = 0;
UINT    nBackgroundH = 0;
UINT    nLcdX = 0;
UINT    nLcdY = 0;
UINT    nLcdZoom = 1;
HDC     hLcdDC = NULL;
HDC     hMainDC = NULL;

static LPBYTE  pbyLcd;

static HBITMAP hLcdBitmap;
static HBITMAP hMainBitmap;

static UINT uLcdTimerId = 0;

static CONST DWORD dwRowTable[] =			// addresses of display row memory
{
	0x1E0, 0x1F0, 0x200, 0x210, 0x220, 0x230, 0x240, 0x250,
	0x260, 0x270, 0x280, 0x290, 0x2A0, 0x2B0, 0x2C0, 0x2D0,
	0x2D8, 0x2C8, 0x2B8, 0x2A8, 0x298, 0x288, 0x278, 0x268,
	0x258, 0x248, 0x238, 0x228, 0x218, 0x208, 0x1F8, 0x1E8
};

static DWORD dwKMLColor[64] =				// color table loaded by KML script
{
	W,B,B,B,B,B,B,B,B,B,B,B,B,B,B,B,
	B,B,B,B,B,B,B,B,B,B,B,B,B,B,B,B,
	I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,
	I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,I
};

static struct
{
	BITMAPINFOHEADER Lcd_bmih;
	RGBQUAD bmiColors[3];
} bmiLcd =
{
	{sizeof(BITMAPINFOHEADER),0/*x*/,0/*y*/,1,8,BI_RGB,0,0,0,ARRAYSIZEOF(bmiLcd.bmiColors),0}
};

VOID UpdateContrast(BYTE byContrast)
{
	// when display is off use contrast 0
	if ((Chipset.IORamM[DSPCTL] & DON) == 0) byContrast = 0;

	bmiLcd.bmiColors[0] = *(RGBQUAD*)&dwKMLColor[byContrast+32]; // pixel off color
	bmiLcd.bmiColors[1] = *(RGBQUAD*)&dwKMLColor[byContrast];    // pixel on  color
	bmiLcd.bmiColors[2] = *(RGBQUAD*)&dwKMLColor[0];			 // background color

	// if pixel off color is undefined, use color 0 for compatibility
	if (I == *(DWORD*)&bmiLcd.bmiColors[0]) bmiLcd.bmiColors[0] = *(RGBQUAD*)&dwKMLColor[0];

	// update palette information
	EnterCriticalSection(&csGDILock);		// asynchronous display update!
	{
		_ASSERT(hLcdDC);
		SetDIBColorTable(hLcdDC,0,ARRAYSIZEOF(bmiLcd.bmiColors),bmiLcd.bmiColors);
	}
	LeaveCriticalSection(&csGDILock);
	UpdateAnnunciators();					// adjust annunciator color
	return;
}

VOID SetLcdColor(UINT nId, UINT nRed, UINT nGreen, UINT nBlue)
{
	dwKMLColor[nId&0x3F] = ((nRed&0xFF)<<16)|((nGreen&0xFF)<<8)|(nBlue&0xFF);
	return;
}

VOID GetSizeLcdBitmap(INT *pnX, INT *pnY)
{
	*pnX = *pnY = 0;						// unknown

	if (hLcdBitmap)
	{
		*pnY = -bmiLcd.Lcd_bmih.biHeight;
		*pnX = LCD_WIDTH * nLcdZoom;
	}
	return;
}

VOID CreateLcdBitmap(VOID)
{
	// create LCD bitmap
	bmiLcd.Lcd_bmih.biWidth = LCD_ROW * nLcdZoom;
	bmiLcd.Lcd_bmih.biHeight = -LCD_HEIGHT * nLcdZoom;
	_ASSERT(hLcdDC == NULL);
	hLcdDC = CreateCompatibleDC(hWindowDC);
	_ASSERT(hLcdDC != NULL);
	hLcdBitmap = CreateDIBSection(hWindowDC,(BITMAPINFO*)&bmiLcd,DIB_RGB_COLORS,(VOID **)&pbyLcd,NULL,0);
	_ASSERT(hLcdBitmap != NULL);
	hLcdBitmap = SelectObject(hLcdDC,hLcdBitmap);
	_ASSERT(hPalette != NULL);
	SelectPalette(hLcdDC, hPalette, FALSE);	// set palette for LCD DC
	RealizePalette(hLcdDC);					// realize palette
	UpdateContrast(Chipset.contrast);
	return;
}

VOID DestroyLcdBitmap(VOID)
{
	// set contrast palette to startup colors
	WORD i = 0;   dwKMLColor[i++] = W;
	while(i < 32) dwKMLColor[i++] = B;
	while(i < 64) dwKMLColor[i++] = I;

	if (hLcdDC != NULL)
	{
		// destroy LCD bitmap
		DeleteObject(SelectObject(hLcdDC,hLcdBitmap));
		DeleteDC(hLcdDC);
		hLcdDC = NULL;
		hLcdBitmap = NULL;
	}
	return;
}

BOOL CreateMainBitmap(LPCTSTR szFilename)
{
	HPALETTE hAssertPalette;

	_ASSERT(hWindowDC != NULL);
	hMainDC = CreateCompatibleDC(hWindowDC);
	_ASSERT(hMainDC != NULL);
	if (hMainDC == NULL) return FALSE;		// quit if failed
	hMainBitmap = LoadBitmapFile(szFilename);
	if (hMainBitmap == NULL)
	{
		DeleteDC(hMainDC);
		hMainDC = NULL;
		return FALSE;
	}
	hMainBitmap = SelectObject(hMainDC,hMainBitmap);
	_ASSERT(hPalette != NULL);
	hAssertPalette = SelectPalette(hMainDC,hPalette,FALSE);
	_ASSERT(hAssertPalette != NULL);
	RealizePalette(hMainDC);
	return TRUE;
}

VOID DestroyMainBitmap(VOID)
{
	if (hMainDC != NULL)
	{
		// destroy Main bitmap
		DeleteObject(SelectObject(hMainDC,hMainBitmap));
		DeleteDC(hMainDC);
		hMainDC = NULL;
		hMainBitmap = NULL;
	}
	return;
}

//****************
//*
//* LCD functions
//*
//****************

static __inline VOID WritePixelQuad(BYTE *p, BYTE a)
{
	if (nLcdZoom == 4)
	{
		CONST DWORD dwPattern[] = { 0x00000000, 0x01010101, 0x02020202 };

		_ASSERT((DWORD) p % sizeof(DWORD) == 0); // check alignment for ARM CPU

		_ASSERT(a < ARRAYSIZEOF(dwPattern));
		*(DWORD *)&p[0*LCD_ROW] = *(DWORD *)&p[4*LCD_ROW]  =
		*(DWORD *)&p[8*LCD_ROW] = *(DWORD *)&p[12*LCD_ROW] = dwPattern[a];
		return;
	}
	if (nLcdZoom == 3)
	{
		p[0*LCD_ROW+0] = p[0*LCD_ROW+1] = p[0*LCD_ROW+2] =
		p[3*LCD_ROW+0] = p[3*LCD_ROW+1] = p[3*LCD_ROW+2] =
		p[6*LCD_ROW+0] = p[6*LCD_ROW+1] = p[6*LCD_ROW+2] = a;
		return;
	}
	if (nLcdZoom == 2)
	{
		CONST WORD wPattern[] = { 0x0000, 0x0101, 0x0202 };

		_ASSERT((DWORD) p % sizeof(WORD) == 0); // check alignment for ARM CPU

		_ASSERT(a < ARRAYSIZEOF(wPattern));
		*(WORD *)&p[0*LCD_ROW] =
		*(WORD *)&p[2*LCD_ROW] = wPattern[a];
		return;
	}
	if (nLcdZoom == 1)
	{
		p[0*LCD_ROW] = a;
		return;
	}
	return;
}

VOID UpdateMainDisplay(VOID)
{
	DWORD d;
	INT   i;
	BYTE  *p,c;

	// illegal zoom factor
	_ASSERT(nLcdZoom >= 1 && nLcdZoom <= 4);

	// on chip with IAM bit set the LCD pins LC[91:60] are ROW line drivers
	// this isn't emulated/simulated so far

	if (!(Chipset.IORamM[DSPCTL]&DON))
	{
		ZeroMemory(pbyLcd, LCD_ROW * nLcdZoom * -bmiLcd.Lcd_bmih.biHeight);
	}
	else
	{
		// next memory position in LCD bitmap
		INT nNextLine = LCD_ROW * nLcdZoom * nLcdZoom;

		// scan complete display area of master
		for (d = MDISP_BEGIN; d <= MDISP_END; d += ARRAYSIZEOF(dwRowTable) / 4)
		{
			// calculate column memory position in LCD bitmap
			p = pbyLcd + ((SDISP_END-SDISP_BEGIN+1+d) >> 3) * nLcdZoom;

			// build display row base on the display ROW table
			for (i = 0; i < ARRAYSIZEOF(dwRowTable); ++i)
			{
				// set pixel state
				c = (*(QWORD *) &Chipset.IORamM[d] & *(QWORD *) &Chipset.IORamM[dwRowTable[i]]) != 0;

				// display row is switched off
				if (*(QWORD *) &Chipset.IORamM[dwRowTable[i]] == 0)
					c = 2;					// use background color

				// write pixel zoom independent
				WritePixelQuad(p,c);

				p += nNextLine;				// next memory position in LCD bitmap
			}
		}

		// scan complete display area of slave
		for (d = SDISP_BEGIN; d <= SDISP_END; d += ARRAYSIZEOF(dwRowTable) / 4)
		{
			// calculate column memory position in LCD bitmap
			p = pbyLcd + ((d-SDISP_BEGIN) >> 3) * nLcdZoom;

			// build display row base on the display ROW table
			for (i = 0; i < ARRAYSIZEOF(dwRowTable); ++i)
			{
				// set pixel state
				c = (*(QWORD *) &Chipset.IORamS[d] & *(QWORD *) &Chipset.IORamM[dwRowTable[i]]) != 0;

				// display row is switched off
				if (*(QWORD *) &Chipset.IORamM[dwRowTable[i]] == 0)
					c = 2;					// use background color

				// write pixel zoom independent
				WritePixelQuad(p,(BYTE)(Chipset.bLcdSyncErr ? 1 : c));

				p += nNextLine;				// next memory position in LCD bitmap
			}
		}
	}

	EnterCriticalSection(&csGDILock);		// solving NT GDI problems
	{
		// using StretchBlt() instead of BitBlt() as workaround for WineHQ bug
		StretchBlt(hWindowDC, nLcdX, nLcdY, LCD_WIDTH * nLcdZoom, -bmiLcd.Lcd_bmih.biHeight,
				   hLcdDC, 0, 0, LCD_WIDTH * nLcdZoom, -bmiLcd.Lcd_bmih.biHeight, SRCCOPY);
		GdiFlush();
	}
	LeaveCriticalSection(&csGDILock);
	return;
}

VOID UpdateAnnunciators(VOID)
{
	const DWORD dwAnnAddr[] = { SLA_ALL, SLA_HALT, SLA_SHIFT, SLA_ALPHA, SLA_BUSY, SLA_BAT, SLA_RAD, SLA_PRINTER };

	INT  i,j,nCount;

	BOOL bAnnOn = (Chipset.IORamM[DSPCTL] & DON) != 0 && (Chipset.IORamS[DSPCTL] & DON) != 0;

	// check all annuncators
	for (i = 1; i < ARRAYSIZEOF(dwAnnAddr); ++i)
	{
		nCount = 0;

		if (bAnnOn)							// annunciators on
		{
			DWORD dwAnn = Npack(Chipset.IORamS+dwAnnAddr[i],8) ^ Npack(Chipset.IORamS+dwAnnAddr[0],8);

			// count the number of set bits
			for (;dwAnn != 0; dwAnn >>= 1)
			{
				if ((dwAnn & 0x1) != 0) ++nCount;
			}
		}

		// contrast table entry of annunciator
		j = Chipset.contrast + nCount - 19;
		if (j < 0)   j = 0;
		if (j >= 31) j = 31;

		DrawAnnunciator(i, nCount > 0, dwKMLColor[j]);
	}
	return;
}

static VOID CALLBACK LcdProc(UINT uEventId, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2)
{
	EnterCriticalSection(&csLcdLock);
	{
		if (uLcdTimerId)					// display update task still active
		{
			UpdateMainDisplay();
		}
	}
	LeaveCriticalSection(&csLcdLock);
	return;
	UNREFERENCED_PARAMETER(uEventId);
	UNREFERENCED_PARAMETER(uMsg);
	UNREFERENCED_PARAMETER(dwUser);
	UNREFERENCED_PARAMETER(dw1);
	UNREFERENCED_PARAMETER(dw2);
}

VOID StartDisplay(VOID)
{
	if (uLcdTimerId)						// LCD update timer running
		return;								// -> quit

	if (Chipset.IORamM[DSPCTL]&DON)			// display on?
	{
		UpdateAnnunciators();				// switch on annunciators
		VERIFY(uLcdTimerId = timeSetEvent(DISPLAY_FREQ,0,(LPTIMECALLBACK)&LcdProc,0,TIME_PERIODIC));
	}
	return;
}

VOID StopDisplay(VOID)
{
	if (uLcdTimerId == 0)					// timer stopped
		return;								// -> quit

	timeKillEvent(uLcdTimerId);				// stop display update
	uLcdTimerId = 0;						// set flag display update stopped

	EnterCriticalSection(&csLcdLock);
	{
		ZeroMemory(pbyLcd, bmiLcd.Lcd_bmih.biWidth * -bmiLcd.Lcd_bmih.biHeight);
	}
	LeaveCriticalSection(&csLcdLock);
	InvalidateRect(hWnd,NULL,FALSE);
	return;
}

VOID ResizeWindow(VOID)
{
	RECT rectWindow;
	RECT rectClient;

	if (hWnd == NULL) return;				// return if window closed

	rectWindow.left   = 0;
	rectWindow.top    = 0;
	rectWindow.right  = nBackgroundW;
	rectWindow.bottom = nBackgroundH;
	AdjustWindowRect(&rectWindow, WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_OVERLAPPED, TRUE);
	SetWindowPos(hWnd, bAlwaysOnTop ? HWND_TOPMOST : HWND_NOTOPMOST, 0, 0,
		rectWindow.right  - rectWindow.left,
		rectWindow.bottom - rectWindow.top,
		SWP_NOMOVE);
	GetClientRect(hWnd, &rectClient);
	AdjustWindowRect(&rectClient, WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_OVERLAPPED, TRUE);
	if (rectClient.bottom < rectWindow.bottom)
	{
		rectWindow.bottom += (rectWindow.bottom - rectClient.bottom);
		SetWindowPos (hWnd, NULL, 0, 0,
			rectWindow.right  - rectWindow.left,
			rectWindow.bottom - rectWindow.top,
			SWP_NOMOVE | SWP_NOZORDER);
	}

	_ASSERT(hWindowDC);						// move destination window
	SetWindowOrgEx(hWindowDC, nBackgroundX, nBackgroundY, NULL);
	InvalidateRect(hWnd,NULL,TRUE);
	return;
}
